"""Handle generic pygame collision.

Maybe this shouldn't ever move a sprite, but always return
a coordinate and any sprites it collides with.

"""

import pygame


# TODO: this is pretty unnecessary now...
def collides_rect(sprite, sprite_group):
    """

    Boilerplate for checking rectangular collision of provided
    sprite against sprites in the provided sprite_group.

    """

    return pygame.sprite.spritecollide(sprite, sprite_group, False)


def collides_rect_mask(sprite, sprite_group):
    """See if provided sprite collides with any in the sprite_group
    by rect first, then check by mask if exists.

    Arguments:
        sprite (pygame.Sprite): ...
        sprite_group (pygame.sprite.Group): See if this ColliderSprite
            collides with any sprites in `sprite_group`.

    Returns:
        None: If there are no collisions.
        pygame.sprite.Sprite: the first sprite to collide with
            this ColliderSprite based on rect and possibly also
            mask (if exists on both this ColliderSprite and the
            other sprite being checked!).

    """

    for sprite_whose_rect_collides in collides_rect(sprite, sprite_group):

        if (hasattr(sprite, 'mask')
                and hasattr(sprite_whose_rect_collides, 'mask')
                and (pygame.sprite
                     .collide_mask(sprite_whose_rect_collides, sprite))):

            return sprite_whose_rect_collides

        elif (not hasattr(sprite, 'mask')) and (not hasattr(sprite, 'mask')):
            return sprite_whose_rect_collides

    return None


def move_as_close_as_possible(sprite, destination, sprite_group):
    """Return how close sprite can go to destination without collision,
    along with the first sprite blocking its progress (if any).

    "Position" herein will always refer to a position
    for sprite.rect.topleft.

    Say you have a SPACESHIP (S), "blank space" (-), the
    destination (D), and a collidable BLOCK (B). If the
    SPACESHIP moves at a velocity of x+5 y+5, this function
    would would take (6, 6) as the `destination`, and it would
    return ((3, 3), <sprite of the BLOCK>). This directly
    mitigates "bad" collision where you could simply jump from
    1,1 to 6,6 despite there being a collision in the path (BLOCK).

        123456
      1 S-----
      2 ------
      3 ------
      4 ---B--
      5 ------
      6 -----D

    In the example above, if there were no BLOCK (B) the return value
    would be ((6,6), None).

    Warning:
        This isn't fast! It lacks any decent heuristics, such as
        line-based collisions. I'll be benchmarking this method
        against line-based collision heuristics in the future.

    Arguments:
        sprite (pygame.Sprite): This sprite is used to incrementally
            move toward the destination, continuously checking
            if colliding with any in sprite_group. The sprite's position
            will not be affected (you will have to update the
            sprite.rect.topleft respectively, yourself).
        destination (tuple[x, y]): The goal coordinate to move to, or
            at least as close to as possible before colliding.
        sprite_group (pygame.sprite.Group): Pygame sprite group, whose
            sprites are check each time we move one pixel
            toward the destination.

    Returns:
        tuple: The first element is the "topleft" coordinate
            representing the closest sprite may move toward the
            destination before a collision occurs. The first element
            coordinate will be one of the following:
              * At the original "topleft" position of sprite, i.e.,
                sprite.rect.topleft
              * The destination provided: if there were no collisions
                moving between original position and destination.
              * In between original and destination positions: when
                there was a collision, this will be the last value
                that didn't collide along that path toward destination.
            The second element is the first sprite which
            prevents progressing further toward destination. The second
            element could be None if sprite can move to destination
            without any collisions from sprite_group.

    """

    original_position = sprite.rect.topleft

    # Figure out the x and y increments!
    #
    # I use "increment" herein to mean "step which approaches,
    # by one, the destination.
    #
    # For both the x and y axis, get the "step"
    # (increment or decrement) which will
    # eventually bring us to the goal.
    goal_x, goal_y = destination

    if goal_x > sprite.rect.left:
        x_increment = 1
    elif goal_x < sprite.rect.left:
        x_increment = -1
    else:
        x_increment = 0

    if goal_y > sprite.rect.top:
        y_increment = 1
    elif goal_y < sprite.rect.top:
        y_increment = -1
    else:
        y_increment = 0

    # This loop will return the first sprite it finds to collide,
    # or we'll return None (later) if it can't find anything!
    while sprite.rect.topleft != (goal_x, goal_y):
        # last_safe_topleft allows us to reset to the last known
        # good/noncolliding coordinate in the event of a collision
        last_safe_topleft = sprite.rect.topleft

        # If y is already at its goal, there's no need to increment
        if sprite.rect.top != goal_y:
            sprite.rect.top += y_increment

        # ... same thing for x and its goal.
        if sprite.rect.left != goal_x:
            sprite.rect.left += x_increment

        colliding_with = collides_rect_mask(sprite, sprite_group)

        if colliding_with:
            sprite.rect.topleft = original_position
            return (last_safe_topleft, colliding_with)

    return (destination, None)


def sprites_in_orthogonal_path(sprite, new_coord, sprite_group):
    """Return the sprites this ColliderSprite would "run through"
    and thus collide with if it moved to new_coord.

    Warning:
        This does not work diagonally! This is only good for
        quickly checking collisions along an orthogonal path.

    Arguments:
        new_coord (tuple[int, int]): topleft coordinate value this
            ColliderSprite would hypotherically have at the end of
            this path.
        sprite_group (pygame.sprite.Group): ...

    """

    current_rect = sprite.rect
    future_rect = sprite.rect.copy()
    future_rect.topleft = new_coord
    collision_rect = future_rect.union(current_rect)
    sprite.rect = collision_rect
    colliding_with = collides_rect(sprite, sprite_group)
    sprite.rect = current_rect
    return colliding_with


def collides_line(sprite, line_point_a, line_point_b, sprite_group):
    """Efficiently check if any sprites along a line collide,
    return True on first result, False if no collision.

    Arguments:
        line_point_a (tuple[int, int]): --

    """

    pass


def lines_intersection(line_a, line_b):
    """Return the point in which lines intersect, else
    return None.

    Arguments:
        line_a (tuple[int, int]): --

    http://webcache.googleusercontent.com/search?q=cache:Ur-EPX41x00J:devmag.org.za/2009/04/17/basic-collision-detection-in-2d-part-2/+&cd=1&hl=en&ct=clnk&gl=us&client=ubuntu

    """

    pass



"""Painless animated sprites which are pygame.sprite.Sprite objects.

Animated sprites from GIF, and all YOU have to know is to `update_state`!
You just have to call `update_state` once per loop with the timedelta:

  >>> AnimatedSprite.from_gif('example.gif')  # doctest: +SKIP
  >>> timedelta = clock.get_time()  # doctest: +SKIP
  >>> animated_sprite.update_state(timedelta)  # doctest: +SKIP

Treat it like a normal pygame sprite, because it is! It's that easy!

A word on `rect` and `mask`:
    In pygame, `Sprite` has the `rect` attribute and optionally the
    `mask` attribute. Both `rect` and `mask` sprite attributes are the
    primary data used in collision detection.

    In this implementation, `Frame` and `AnimatedSprite` both have the
    `mask` and `rect` attributes. Don't use `rect for positional stuff,
    e.g., don't do something like `animatedsprite.rect.topleft = (11, 22)`.
    The `mask` and `rect` of an `AnimatedSprite` are merely a reference to
    the current `Frame`'s `rect` and `mask`! When you load an animated
    sprite, you can specify a threshold:

      >>> AnimatedSprite.from_gif('example.gif',
      ...                         mask_threshold=254)  # doctest: +SKIP

    Whereas, the threshold denotes the alpha transparency value for a
    pixel, which is opaque enough to be "set".

What's supported:
    Right now there's only support for GIFs, but I'd like to expand that
    (especially since GIFs can't really take advantage of alpha
    transparency).

"""

import pygame
from PIL import Image


# NOTE: could be a sprite...
class Frame(object):
    """A frame of an AnimatedSprite animation.

    Attributes:
        surface (pygame.Surface): The pygame image which is used
            for a frame of an animation.
        mask (pygame.Mask): Mask automatically generated from the
            supplied surface (see above). Only exists if mask_threshold
            was >0 on init.
        duration (integer): Milliseconds this frame lasts. How
            long this frame is displayed in corresponding animation.
        start_time (integer): The animation position in milleseconds,
            when this frame will start being displayed.

    See Also:
        * AnimatedSprite.frames_from_gif()
        * AnimatedSprite.animation_position

    """

    def __init__(self, surface, start_time, duration, mask_threshold=0):
        """Create a frame using a pygame surface, the start time,
        and the duration time.

        Args:
            surface (pygame.Surface): The surface/image for this
                frame.
            start_time (int): Millisecond this frame starts. This
                frame is a part of a larger series of frames and
                in order to render the animation properly we
                need to know when each frame begins to be drawn,
                while duration signifies when it ends.
            duration (integer): Milleseconds this frame lasts. See:
                start_time argument description.
            mask_threshold (int): Valid values 0-254. Alpha values
                ABOVE this provided number are marked as "solid"/
                collidable/set. If this is not greater than zero,
                the mask is not generated.

        """

        self.surface = surface
        self.duration = duration
        self.start_time = start_time
        self.end_time = start_time + duration

        if mask_threshold > 0:
            self.mask = pygame.mask.from_surface(surface, mask_threshold)

    def __repr__(self):
        s = "<Frame duration(%s) start_time(%s) end_time(%s)>"

        return s % (self.duration, self.start_time, self.end_time)


class AnimatedSprite(pygame.sprite.Sprite):
    """Animated sprite with mask, loaded from GIF.

    Supposed to be mostly uniform with the Sprite API.

    Attributes:
        total_duration (int): The total duration of of this
            animation in milliseconds.
        image (pygame.Surface): Current surface belonging to
            the active frame. Set once per tick through
            the AnimatedSprite.update_state() method.
        rect (pygame.Rect): Does not reflect position, only
            area. Updated once per tick, to reflect current
            frame's rect, in AnimatedSprite.update_state().
        mask (pygame.Mask): If the first frame had a mask
            attribute, then we assume all do, and this
            (optional) attribute points to the active
            frame's mask attribute.
        active_frame_index (int): Frame # which is being
            rendered/to be rendered. Also updated once per
            tick, see the AnimatedSprite.update_state() method.
        active_frame: The current surface representing this
            animation at its current animation position. Set
            once per tick through the update_state() method.
        animation_position (int): Animation position in
            milliseconds; milleseconds elapsed in this
            animation. This is used for determining
            which frame to select. Set once per tick through
            the AnimatedSprite.update_state() method.

    See Also:
        * :class:`pygame.sprite.Sprite`
        * :class:`Frame`

    """

    def __init__(self, frames):
        """Create this AnimatedSprite using
        a list of Frame instances.

        Args:
            frames (list[Frame]): A properly assembled list of frames,
                which assumes that each Frame's start_time is greater
                than the previous element and is the previous element's
                start time + previous element/Frame's duration. Here
                is an example of aformentioned:

                >>> frame_one_surface = pygame.Surface((16, 16))
                >>> frame_one = Frame(frame_one_surface, 0, 100)
                >>> frame_two_surface = pygame.Surface((16, 16))
                >>> frame_two = Frame(frame_two_surface, 100, 50)

        Note:
            In the future I may add a method for verifying the
            validity of Frame start_times and durations.

        """

        super(AnimatedSprite, self).__init__()
        self.frames = frames
        self.total_duration = self.get_total_duration(self.frames)
        self.active_frame_index = 0
        self.active_frame = self.frames[self.active_frame_index]

        # animation position in milliseconds
        self.animation_position = 0

        # this gets updated depending on the frame/time
        # needs to be a surface.
        self.image = self.frames[0].surface

        # represents the animated sprite's position
        # on screen.
        self.rect = self.image.get_rect()

        # making the bold assumption that if the
        # first frame has a mask, as do the rest.
        if hasattr(self.frames[0], 'mask'):
            self.mask = self.frames[0].mask

    def __getitem__(self, frame_index):
        """Return the frame corresponding to
        the supplied frame_index.

        Args:
            frame_index (int): Index number to lookup
                a frame by element number in the
                self.frames list.

        Returns:
            Frame: The frame of this animation at the
                specified index of frame_index.

        """

        return self.frames[frame_index]

    def largest_frame_size(self):
        """Return the largest frame's (by area)
        dimensions as tuple(int x, int y).

        Returns:
            tuple (x, y): pixel dimensions of the largest
                frame surface in this AnimatedSprite.

        """

        largest_frame_size = (0, 0)

        for frame in self.frames:
            largest_x, largest_y = largest_frame_size
            largest_area = largest_x * largest_y

            frame_size = frame.surface.get_size()
            frame_x, frame_y = frame_size
            frame_area = frame_x * frame_y

            if frame_area > largest_area:
                largest_frame_size = (frame_size)

        return largest_frame_size

    @classmethod
    def from_gif(cls, path_or_readable, mask_threshold=0):
        """The default is to create from gif bytes, but this can
        also be done from other methods...

        Create a list of surfaces (frames) and a list of their
        respective frame durations from an animated GIF.

        Args:
            path_or_readable (str|file-like-object): Either a string
                or an object with a read() method. So, either a path
                to an animated GIF, or a file-like-object/buffer of
                an animated GIF.
            mask_threshold (int): An optional keyword argument which
                must be >0 to generate masks automatically per frame.
                This value is used to note which parts are opaque and
                thus collidable, and which values are not. Think of
                RGBA, valid values are 0-254. See also: Frame().

        Returns:
            AnimatedSprite: --

        """

        pil_gif = Image.open(path_or_readable)

        frame_index = 0
        frames = []
        time_position = 0

        try:

            while True:
                duration = pil_gif.info['duration']
                frame_sprite = cls.pil_image_to_pygame_surface(pil_gif)
                frame = Frame(surface=frame_sprite,
                              start_time=time_position,
                              duration=duration,
                              mask_threshold=mask_threshold)
                frames.append(frame)
                frame_index += 1
                time_position += duration
                pil_gif.seek(pil_gif.tell() + 1)

        except EOFError:

            pass  # end of sequence

        return AnimatedSprite(frames)

    def update(self, timedelta):
        """Manipulate the state of this AnimatedSprite, namely
        the on-screen/viewport position (not absolute) and
        using the timedelta to do animation manipulations.

        Using the game's timedelta we decipher the animation position,
        which in turn allows us to locate the correct frame.

        Sets the image attribute to the current frame's image. Updates
        the rect attribute to the new relative position and frame size.

        Warning:
            Since we're changing the rect size on-the-fly, this can
            get the player stuck in certain boundaries. I will be
            remedying this in the future.

        Args:
            timedelta (int|float): Typically from the game
                clock (pygame.time.Clock) via clock.get_time().
                Used to update the animation position.

        """

        self.animation_position += timedelta

        if self.animation_position >= self.total_duration:
            self.animation_position = (self.animation_position %
                                       self.total_duration)
            self.active_frame_index = 0

        while (self.animation_position >
               self.frames[self.active_frame_index].end_time):

            self.active_frame_index += 1

        # NOTE: the fact that I'm using -1 here seems sloppy/hacky
        self.image = self.frames[self.active_frame_index - 1].surface

        self.rect.size = self.image.get_size()
        self.active_frame = self.frames[self.active_frame_index]

        # if we have a mask, let's update our pointer!
        # again, we make the bold assumption that if
        # our first frame had a mask, then the rest do.
        if hasattr(self, 'mask'):
            self.mask = self.active_frame.mask

    @staticmethod
    def get_total_duration(frames):
        """Return the total duration of the animation in milliseconds,
        milliseconds, from animation frame durations.

        Args:
            frames (List[AnimatedSpriteFrame]): --

        Returns:
            int: The sum of all the frame's "duration" attribute.

        """

        return sum([frame.duration for frame in frames])

    @staticmethod
    def pil_image_to_pygame_surface(pil_image):
        """Convert PIL Image() to RGBA pygame Surface.

        Args:
            pil_image (Image): image to convert to pygame.Surface().

        Returns:
            pygame.Surface: the converted image

        Example:
            >>> from PIL import Image
            >>> gif = Image.open('demo/test_scene/test.gif')
            >>> AnimatedSprite.pil_image_to_pygame_surface(gif)
            <Surface(10x10x32 SW)>

        """

        image_as_string = pil_image.convert('RGBA').tobytes()

        return pygame.image.fromstring(image_as_string,
                                       pil_image.size,
                                       'RGBA')



# TODO: module level docstring
import sys
import pygame

PY3 = sys.version_info[0] == 3
range = range if PY3 else xrange


class SurfaceLayers(object):
    """Ordered series of pygame surfaces, each the size of the target
    surface given at creation time.

    Arguments:
        target_surface (pygame.Surface): Surface that will have
            have the layers blitted to when render() is called. The size
            of this surface is used as the size of the generated layers.
        number_of_layers (int): Number of layers to generate.

    """

    def __init__(self, target_surface, number_of_layers):
        self._target_surface = target_surface
        self._surface_layers = self.create_surface_layers(target_surface,
                                                          number_of_layers)

    @staticmethod
    def create_surface_layers(target_surface, number_of_layers):
        """Create a list of pygame surfaces
        the size of the target surface.

        Arguments:
            target_surface (pygame.Surface): The surface
                whose dimensions will be used for each layer.
            number_of_layers (int): The number of surfaces to
                create/return.

        Returns:
            list[pygame.Surface]: List of surfaces

        """

        surface_layers = []

        for i in range(number_of_layers):
            surface = pygame.surface.Surface(target_surface.get_size(),
                                             pygame.SRCALPHA, 32)
            surface_layers.append(surface)

        return surface_layers

    def __getitem__(self, key):
        """Access a surface by z-index.

        Arguments:
            key (int): The z-index of the surface.

        Raises:
            IndexError: When the z-index is invalid.

        Returns:
            pygame.Surface: The surface belonging
                to the z-index specified.

        """

        return self._surface_layers[key]

    def __len__(self):
        """Return the number of layers.

        Returns:
            int: Number of members in self.surface_layers.

        """

        return len(self._surface_layers)

    def __iter__(self):
        """Iterate through the surface layers.

        Yields:
            pygame.surface.Surface

        """

        for surface in self._surface_layers:
            yield surface

    def render(self):
        """Draw each layer onto the target surface in the correct order.

        We're wiping surface layers because there may be a layer higher up
        which has a lot of transparent air but parts of pillars or whatever;
        if we didn't wipe you'd see that column drag across the screen as the
        camera moved.

        """

        for surface in self._surface_layers:
            self._target_surface.blit(surface, (0, 0))

        target_surface = self._target_surface
        number_of_layers = len(self._surface_layers)
        self._surface_layers = self.create_surface_layers(target_surface,
                                                          number_of_layers)



"""Camera control on a surface.

Camera is just a pygame surface, so treat it as such!

A camera allows you to have a viewport which is smaller than the size
of the surface you're hoping to pan that viewport around on. Think
about looking at a big painting through a paper towel tube.

Cameras have an output resolution it scales its viewport to, plus a
behavior which governs how a Camera pans on its associated surface.

"""

import pygame


class CameraOutOfBounds(Exception):
    """The subsurface, which is the camera's view,
    goes beyond the surface the camera's viewing.

    Camera has been moved so that its area exceeds
    the surface it uses to make a subsurface.

    Attributes:
        camera (Camera):

    """

    def __init__(self, camera):
        super(CameraOutOfBounds, self).__init__(camera)
        self.camera = camera


class CameraBehavior(object):
    """How a camera moves. How it handles boundaries,
    character movement, etc.

    This CameraBehavior, the default, keeps the focal rectangle
    in the top left of the camera view.

    You'll want to inherit this class when creating
    a CameraBehavior, including overriding the
    move method.

    """

    @staticmethod
    def move(camera, focal_rectangle):
        """Move the camera, keeping the focal rectangle in
        the top left of the camera view.

        This method should be overridden in a child class.

        Arguments:
            camera (sappho.camera.Camera): Associated
                Sappho camera object to control.
            focal_rectangle (pygame.Rect): Rectangle which
                is used to possibly adjust camera position.

        """

        scroll_position = focal_rectangle.topleft
        camera.view_rect.topleft = scroll_position


class CameraCenterBehavior(CameraBehavior):
    """A camera behavior that centers the
    focal rectangle on the screen.

    Will not cause Camera.update_state() to raise
    CameraOutOfBounds, because the move logic
    prevents such from occuring!

    """

    @staticmethod
    def move(camera, focal_rectangle):
        """Move the camera, keeping the focal rectangle
        in the center of the screen where possible.

        Arguments:
            camera (sappho.camera.Camera):
            focal_rectangle (pygame.Rect): Rectangle which
                is used to possibly adjust camera position.

        """

        new_view_rect = camera.view_rect.copy()
        new_view_rect.center = focal_rectangle.center

        # Make sure the camera isn't centered in a way that
        # exceeds the environment's dimensions (don't go off
        # the map!)
        if new_view_rect.left < 0:
            new_view_rect.left = 0

        if new_view_rect.top < 0:
            new_view_rect.top = 0

        if new_view_rect.bottom > camera.source_resolution[1]:
            new_view_rect.bottom = camera.source_resolution[1]

        if new_view_rect.right > camera.source_resolution[0]:
            new_view_rect.right = camera.source_resolution[0]

        camera.view_rect = new_view_rect


class Camera(pygame.surface.Surface):
    """Surface that acts as a scrollable view, with optional scaling
    onto another surface.

    Attributes:
        source_resolution (tuple[int, int]): Maximum size of the
            environment being portrayed. If you have a map with many
            inconsistently sized layers, this should be the size of
            all of those layers flattened onto a single new layer.
            Anything beyond this size will not be on camera.
        output_resolution (tuple[int, int]): Resolution to scale up the
            view of the surface to
        view_rect (pygame.Rect): Rectangle area of this camera's view,
            which is used to create the subsurface, which is scaled
            to output_resolution and blit to self/camera.
        behavior (CameraBehavior): The initial behavior to use for this
            Camera. The :py:class:`CameraBehavior <sappho.CameraBehavior>`
            that this Camera uses to control movement.

    """

    def __init__(self, source_resolution, output_resolution,
                 view_resolution, behavior=None):

        """Create a Camera!

        Arguments:
            view_resolution (tuple[int, int]): used to create
                view_rect attribute.

        """

        super(Camera, self).__init__(output_resolution)

        self.source_surface = pygame.surface.Surface(source_resolution,
                                                     pygame.SRCALPHA)
        self.source_resolution = source_resolution
        self.output_resolution = output_resolution
        self.view_rect = pygame.Rect((0, 0), view_resolution)
        self.behavior = behavior or CameraBehavior()

    def update_state(self, timedelta):
        """Update the Camera to point to the
        current scroll position.

        Gets the "view" from the current scroll position
        and camera resolution. That view is used to create
        a subsurface of the "source subsurface," scaled to
        the target_resolution. The new subsurface is then
        blit to the camera (which is a surface, itself!).

        Arguments:
            timedelta (None): This isn't used at all, but it
                is included as an argument for consistency.

        """

        try:
            subsurface = self.source_surface.subsurface(self.view_rect)
        except ValueError:
            raise CameraOutOfBounds(self)

        scaled_surface = pygame.transform.scale(subsurface,
                                                self.output_resolution)

        # Blit the scaled surface to this camera (which is also a surface)
        super(Camera, self).blit(scaled_surface, (0, 0))

    def scroll_to(self, focal_rectangle):
        """Scroll to the given focal rectangle using the current behavior.

        Parameters:
            focal_rectangle (pygame.Rect): Rectangle to possibly update
                the view position to using the camera's current behavior

        """

        self.behavior.move(self, focal_rectangle)



#!/usr/bin/env python

"""Sappho package installer.

Distributing:

  $ setup.py sdist bdist_wheel
  $ twine upload dist/sappho-0.2.3.tar.gz dist/sappho-0.2.3*.whl
  $ rm -rf dist

  You'll need the wheel, twine package for bdist_wheel. Don't forget
  to clear your dist when finished.

"""

import sys
from setuptools import setup
from distutils.version import StrictVersion


# Build the list of packages required according to Python version
# Pygame isn't on Pypi.
install_requires = ['Pillow>=2, <4']  # NOTE: why this specific Pillow range?

# x.y.z
python_version = StrictVersion('.'.join(str(n) for n in sys.version_info[:3]))

# the `enum` package is a backport of Python 3.5 enum,
# we only want it in earlier versions of python
if python_version < StrictVersion('3.5'):
    install_requires.append('enum34')

exec(open('sappho/__init__.py').read())
setup(name='sappho',
      packages=['sappho'],
      version=__version__,
      description='2D game engine (pygame)',
      setup_requires=['setuptools-markdown'],
      install_requires=install_requires,
      long_description_markdown_filename='README.md',
      author='Lillian Lemmer',
      author_email='lillian.lemmer@hypatiasoftware.org',
      license='MIT',
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Natural Language :: English',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4',
                   'Topic :: Games/Entertainment :: Role-Playing',
                   'Topic :: Software Development :: Libraries :: pygame',
                  ],
      keywords=('games gaming development sprites adventure game tilemap '
                'tilesheet zelda gamedev 2d')
    )




# TODO: module level docstring
"""Tile engine; tile map, tilesheet system.

Has impassability support.

"""

import sys
import pygame

import xml.etree.ElementTree as ET

PY3 = sys.version_info[0] == 3
range = range if PY3 else xrange

# TODO: this should be configurable in the
# tilesheet rules.
DEFAULT_AUTOMASK_THRESHOLD = 125
"""int: The alpha threshold (0-255) which
a pixel will be marked as NOT SET.
"""


class Flags(object):
    SOLID = "solid"


class Tile(pygame.sprite.Sprite):
    """A tile object is a sprite, which is typically
    a subsurface of a tilesheet.

    Attributes:
        id_ (str): The ID of this tile in its respective
            tilesheet.
        image (pygame.surface.Surface):
        flags (set): a series of strings which
            represent a boolean state of something
            about this tile, e.g., "solid_block"
            or "auto_mask".
        rect (pygame.Rect): Only exists if "solid"
            flag was specified.

    """

    def __init__(self, id_, image, flags=None):
        super(Tile, self).__init__()
        self.id_ = id_
        self.image = image
        self.flags = flags or set([])

        if Flags.SOLID in self.flags:
            self.rect = self.image.get_rect()
            self.mask = pygame.mask.from_surface(image,
                                                 DEFAULT_AUTOMASK_THRESHOLD)

    def copy(self):
        tile = Tile(self.id_, self.image, self.flags)

        if hasattr(self, 'rect') and hasattr(self, 'mask'):
            tile.rect = self.rect.copy()
            tile.mask = self.mask

        return tile


class Tilesheet(object):
    """Efficient place to  update, subsurface tile
    graphics.

    Attributes:
        surface (pygame.surface.Surface):
        tiles (list[Tile]): The tiles composing this
            Tilesheet.
        tile_size (tuple[int, int]): The size of each
            tile in pixels (x, y).

    """

    def __init__(self, surface, tiles, tile_size):
        self.surface = surface
        self.tiles = tiles
        self.tile_size = tile_size

    @staticmethod
    def parse_rules(path_to_rules_file):
        """Get properties for tile IDs from the
        path_to_rules_file.

        A rule file looks something like this:

            83,99,44-77=SOLID,SOMEFLAG
            40,110=SOLID
            10=SOMEFLAG

        So as you can see, you can set multiple tile's properties
        by listing them out and using ranges of tile IDs.

        Argument:
            path_to_rules_file (str): Path to the rules file to parse

        Returns:
            dict: You can lookup a tile by id (key) and get its
                rules (value, set). It should look something like
                this:

                >>> {0: set(['solid']),
                ...  1: set(['solid', 'foo'])}  # doctest: +SKIP

        """

        tile_rules = {}

        with open(path_to_rules_file) as f:
            rules = [line.strip() for line in f.readlines()]

        # For each line, basically
        for rule in rules:
            tile_ids_affected, flags = rule.split('=')
            tile_ids_affected = tile_ids_affected.split(',')
            flags = set([flag.strip() for flag in flags.split(',')])

            for tile_id in tile_ids_affected:

                if '-' in tile_id:
                    # This is a range of tile IDs!
                    first_id, last_id = tile_id.split('-')
                else:
                    # Just one tile ID!
                    first_id = last_id = tile_id

                # Create a dictionary whose keys are
                # first_id through last_id, and values
                # are the flags defined in the rule.
                tile_id_range = range(int(first_id), int(last_id) + 1)
                this_iteration_rules = dict.fromkeys(tile_id_range,
                                                     flags)
                # ... finally updating our overall collection
                # of all the tile rules with the rules found
                # this iteration!
                tile_rules.update(this_iteration_rules)

        return tile_rules

    @classmethod
    def from_file(cls, file_path, tile_width, tile_height):
        """Creates a tilesheet, and its tiles, from a
        file, as well as tile width and height specification.

        Arguments:
            file_path (str): Path to the tilesheet image
            tile_width (int): Width of each tile
            tile_height (int): Height of each tile

        """

        tilesheet_surface = pygame.image.load(file_path)
        tile_rules = cls.parse_rules(file_path + ".rules")

        tile_size = (tile_width, tile_height)
        tilesheet_width, tilesheet_height = tilesheet_surface.get_size()
        tilesheet_width_in_tiles = tilesheet_width // tile_width
        tilesheet_height_in_tiles = tilesheet_height // tile_height
        total_tiles = tilesheet_width_in_tiles * tilesheet_height_in_tiles

        # tile initialization; buid all the tiles
        tiles = []

        for tile_id in range(total_tiles):
            subsurface = cls.tile_subsurface_from_tile_id(tilesheet_surface,
                                                          tile_size,
                                                          tile_id)

            flags = tile_rules[tile_id] if tile_id in tile_rules else set([])
            tile = Tile(id_=tile_id,
                        image=subsurface,
                        flags=flags)
            tiles.append(tile)

        return Tilesheet(tilesheet_surface, tiles, tile_size)

    @staticmethod
    def tile_subsurface_from_tile_id(tilesheet_surface, tile_size, tile_id):
        """Create a subsurface for a tile from the tilesheet surface.

        Arguments:
            tilesheet_surface (pygame.surface.Surface):
            tile_size (list[int, int]): (x, y) pixel dimensions
                of all tiles.
            tile_id (int): The tile ID to graph the graphic from
                the tilesheet surface for.

        Return:
            pygame.surface.Surface:

        """

        tilesheet_width_in_tiles = (tilesheet_surface.get_size()[0] /
                                    tile_size[0])
        top_left_in_tiles = index_to_coord(tilesheet_width_in_tiles,
                                           tile_id)
        subsurface_top_left = (top_left_in_tiles[0] * tile_size[0],
                               top_left_in_tiles[1] * tile_size[1])
        position_rect = pygame.Rect(subsurface_top_left, tile_size)
        area_on_tilesheet = position_rect
        subsurface = tilesheet_surface.subsurface(position_rect)

        return subsurface


class TileMap(object):
    """A 2D grid arrangement of tiles, accompanied by
    passability information.

    This is generally one layer of a larger map, and ideally
    should be blitted to a :class:`sappho.layers.SurfaceLayers` object
    to handle multiple map layers.

    Arguments:
        tilesheet (Tilesheet): Tilesheet to use for this map
        tiles (list[list[Tile]]): List of rows (a row itself is a list),
            where each row contains the appropriate amount of
            :class:`Tile` objects representing the tiles of
            this TileMap.

    """

    def __init__(self, tilesheet, tiles):
        self.tilesheet = tilesheet
        self.tiles = tiles
        self.collision_group = self.set_solid_tiles_topleft(self.tiles)

    def set_solid_tiles_topleft(self, tiles):
        """The rectangles from tiles do not contain positional
        data (all of their toplefts are [0, 0]). This method
        sets the appropriate topleft per rectangle, considering
        the position of the tile in the tilemap.

        Warning:
            Only tiles with the `solid` flag will have their
            topleft set. Keep in mind immutability; the tile
            instances are being modified.

        Returns:
            pygame.sprite.Group: All the collidable tiles which
                actually have rect and mask attributes.

        """

        collidable_tiles_for_sprite_group = []

        for y, row_of_tiles in enumerate(self.tiles):

            for x, tile in enumerate(row_of_tiles):

                # If this flag is present we know the
                # tile will have the rect and mask attribs
                if Flags.SOLID in tile.flags:
                    left_top = (x * self.tilesheet.tile_size[0],
                                y * self.tilesheet.tile_size[1])
                    tile.rect.topleft = left_top
                    collidable_tiles_for_sprite_group.append(tile)

        return pygame.sprite.Group(*collidable_tiles_for_sprite_group)

    def to_surface(self):
        """Blit the TileMap to a surface

        Returns:
            pygame.Surface:
                Surface containing the tilemap

        """

        tile_size_x, tile_size_y = self.tilesheet.tile_size

        width_in_tiles = len(self.tiles[0])
        height_in_tiles = len(self.tiles)

        layer_size = (width_in_tiles * tile_size_x,
                      height_in_tiles * tile_size_y)

        new_surface = pygame.Surface(layer_size, pygame.SRCALPHA, 32)
        new_surface.fill([0, 0, 0, 0])

        for y, row_of_tiles in enumerate(self.tiles):

            for x, tile in enumerate(row_of_tiles):
                # blit tile subsurface onto respective layer
                tile_position = (x * tile_size_x, y * tile_size_y)
                new_surface.blit(tile.image, tile_position)

        return new_surface

    @classmethod
    def from_csv_string_and_tilesheet(cls, csv_string, tilesheet, firstgid=0):
        """Create a tilemap using a CSV of tile IDs and
        a tilesheet.

        Arguments:
            csv_string (str):
            tilesheet (Tilesheet):
            firstgid (int): ID of the first tile

        """

        sheet = []

        for y, line in enumerate(csv_string.split('\n')):
            row = []

            for x, tile_id_string in enumerate(line.split(',')):

                if not tile_id_string:

                    continue

                # This tile needs to be COPIED to be counted
                # as a separate sprite, otherwise if you change
                # the same tile type from tilesheet at one location,
                # it'll change all the other locations of that same
                # tile type.
                tile_id = int(tile_id_string) - firstgid
                tile = tilesheet.tiles[tile_id].copy()
                row.append(tile)

            sheet.append(row)

        return cls(tilesheet, sheet)


def index_to_coord(width, i):
    """Return the 2D position (x, y) which corresponds to 1D index.

    Examples:
      If we have a 2D grid like this:

      0 1 2
      3 4 5
      6 7 8

      We can assert that element 8 is of the coordinate (2, 2):
      >>> (2, 2) == index_to_coord(3, 8)
      True

    """

    if i == 0:

        return (0, 0)

    else:

        return ((i % width), (i // width))


def tmx_file_to_tilemaps(tmx_file_path, tilesheet):
    """Read TMX file from path and return
    a list of TileMaps (one TileMap per layer).

    Arguments:
        tmx_file_path (str)
        tilesheet (Tilesheet)

    Returns:
        list[TileMap]: Each layer gets its own TileMap!

    Note:
        Uses CSV layers; TMX allows all kinds, but
        CSV is the default.

    """

    tree = ET.parse(tmx_file_path)
    root = tree.getroot()  # <map ...>

    firstgid = int(root.findall(".//tileset")[0].attrib["firstgid"])

    tilemaps = []

    for layer_data in root.findall(".//layer/data"):
        data_encoding = layer_data.attrib['encoding']

        if data_encoding != 'csv':

            raise TMXLayersNotCSV(data_encoding)

        layer_csv = layer_data.text.strip()
        layer_tilemap = TileMap.from_csv_string_and_tilesheet(layer_csv,
                                                              tilesheet,
                                                              firstgid)

        tilemaps.append(layer_tilemap)

    return tilemaps



"""Particle system definitions.

There's a few basic concepts here.  Note that all functions affecting physics,
emission, launch, and display are implemented as classes with a `__call__`
magic method, but could just as easily be pure functions since no extra info
is used on these.

    Particle: this represents each basic particle in the system, in particular
        its physical position, velocity, and lifetime.

    Emitter: this determines the number of particles emitted and time between
        particle emission.  It doesn't determine anything about particle
        location speed, etc., as an emitter simply takes a time delta (in
        seconds) and returns the number of particles to emit.

    Launcher: this is the same as a physics function.  Launchers are allowed
        to operate on a new particle (copy of a particle system's example
        particle) for 1 second effective time, instantaneously in real time.
        So an acceleration of 20 pixels / sec / sec, when put in a launcher,
        will give the particle a starting velocity of 20 pixels / sec.

    Physics: each physics function take a time delta and a particle and
        modifies the particle in some way.  It could move or accelerate the
        particle, change the particle's species, etc.

    Artist: this receives a surface to draw on, and a particle, and is
        responsible for drawing the particle on the surface as specified.
        Typically a drawer decides what the particle looks like, if it's
        stretched by velocity, etc.

    ParticleSystem: the current state of a system of particles, including
        all current particles, the physics that affects them, and the artists
        that illustrate them.
"""
from __future__ import division
import copy
import itertools
import math
import random


try:  # pragma: no cover
    import pygame
except ImportError:  # pragma: no cover
    # This allows our particle tests to run even if pygame won't :D
    class pygame:
        BLEND_RGBA_MULT = 8


# The biggest number less than infinity-plus-one
INFINITY = float('inf')

OUT_OF_PARTICLES = -1


# Indicate to center artist's image
def CENTER(image):
    return tuple(x // 2 for x in image.get_size())


class Particle(object):
    """This represents the state of a particle in the system.

    Contains all attributes expected by methods in this module, but
    you're always free to duck-type a particle class with more or fewer
    attributes as you need.  However, note that whatever particle class
    you make should be shallow-copiable by `copy.copy`.

    Attributes:
        x (float): X position of the particle's center, in pixels.
        y (float): Y position of the particle's center,.
        dx (float): X velocity of the particle, in pixels / second.
        dy (float): Y valocity of the particle, in pixels / second.
        life (float): Expected remaining lifetime of the particle, in seconds.
        species (str): Arbitrary name that can affect which physics or
            display to use for the particle.  (This attribute is currently
            unused, and may be used as desired by custom methods.)
        initial_life (float): Life that was initially granted to the particle
            after launch.  This allows one to calculate the remaining
            life fraction of the particle.
    """
    __slots__ = ['x', 'y', 'dx', 'dy', 'life', 'species', 'initial_life']

    def __init__(self, x, y, dx=0, dy=0, life=0, species=None):
        """Just initialize with attributes default values.
        """
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.life = life
        self.species = species
        self.initial_life = life

    def __repr__(self):
        parts = list()
        parts.append(
            'Particle({!r}, {!r}'.format(self.x, self.y)
        )
        if self.dx or self.dy:
            parts.append(', dx={!r}, dy={!r}'.format(self.dx, self.dy))
        parts.append(', life={!r}'.format(self.life))
        if self.species is not None:
            parts.append(', species={!r}'.format(self.species))
        parts.append(')')
        return ''.join(parts)


class ParticleSystem(object):
    """A particle system contains current state of particles, evolution
    functions, and illustration functions.

    After configuring a particle system with `__init__`, users are expected
    to regularly call `update_state(dt)` with time deltas to update physical
    state of the system, & `draw_on(surface)` to draw the particles on a
    pygame surface.  You may also periodically check `is_alive` to determine
    if the particle system has run out of particles and can safely be deleted.
    """
    def __init__(
            self,
            origin_particle,
            emitter,
            launcher=None,
            physics=None,
            artist=None,
            particle_limit=512,
            launcher_dt=1.0,
    ):
        """Create a new particle system.

        Arguments:
            origin_particle (Particle): A particle that defines the origin
                of the system, including its current location & velocity, and
                the initial lifetime remaining of a particle.
            emitter (callable): A function that takes a time delta (in seconds)
                and returns the number of new particles to emit.
            launcher (callable): A physics function that is called on a new
                particle just after being created with an effective time
                delta of one second.  This can apply initial velocity,
                lifetime, etc of the particle.
            physics (callable): A physics function that take a time delta (
                in seconds) and a particle, and applies an update to the
                particle physics that would occur in that period of time.
                Examples include gravitational acceleration, inertia, etc.
            artist (callable): A function taking a surface & a particle,
                that draws the particle however it likes on the given
                surface.
            particle_limit (int): The maximum ever allowed number of particles.
            launcher_dt (float): Effective number of seconds the launcher
                physics is applied to a new particle.

        """
        self.origin = origin_particle
        self.emitter = emitter
        self.launcher = launcher or (lambda *x: None)
        self.launcher_dt = launcher_dt
        self.physics = physics or PhysicsInertia()
        self.is_emitting = True
        self.artist = artist
        self.particle_limit = particle_limit
        self.particles = list()

    def is_alive(self):
        """Is this particle system still running? Either emitting new
        particles, or still illustrating old ones?

        returns (bool): Whether the system is still running.
        """
        return self.particles or self.is_emitting

    def update_state(self, dt):
        """Simulate physics for `dt` seconds.

        Arguments:
            dt (float): Number of elapsed seconds since last update.
        """
        new_particle_count = self._get_new_particle_count(dt)
        for _ in range(new_particle_count):
            self._launch()
        for particle in self.particles:
            self.physics(dt, particle)
        self._discard_dead_particles()

    def draw_on(self, surface):
        """Draw all particles on the given surface."""
        for particle in self.particles:
            self.artist(surface, particle)

    # Internal methods
    def _get_new_particle_count(self, dt):
        """Get count of new particles to create in next `dt` seconds.

        This isn't necessarily the same as the result of just running the
        emitter, because it also applies any particle limit that's in
        effect.
        """
        desired_particle_count = self.emitter(dt)
        self.is_emitting = not _indicates_dead_emitter(desired_particle_count)
        particle_slots_available = self.particle_limit - len(self.particles)
        return min(desired_particle_count or 0, particle_slots_available)

    def _launch(self):
        """Build and launch a particle."""
        new_particle = copy.copy(self.origin)
        self.launcher(self.launcher_dt, new_particle)
        new_particle.initial_life = new_particle.life
        self.particles.append(new_particle)

    def _discard_dead_particles(self):
        """Get rid of all no longer active particles (life <= 0)."""
        write_index = 0
        for read_index, particle in enumerate(self.particles):
            if particle.life > 0:
                if read_index > write_index:
                    self.particles[write_index] = particle
                write_index += 1
        del self.particles[write_index:]


class EmitterComposite(object):
    """Combine several emitters into one."""
    def __init__(self, *emitters):
        """Create composite of multiple emitters.

        Arguments:
            emitters (list of emitter callables): Emitter callables, each of
                which indicates number of particles to emit in `dt` seconds
        """
        self.emitters = list(emitters)
        self._alive_count = len(self.emitters)

    def add(self, *emitters):
        """Add multiple emitters.

        Arguments:
            emitters (list of emitter callables): Emitter callables, each of
                which indicates number of particles to emit in `dt` seconds
        """
        self.emitters.extend(emitters)
        self._alive_count = len(self.emitters)
        return self

    def __call__(self, dt):
        """How many particles to emit in `dt` seconds.

        Arguments:
            dt: Elapsed time in seconds.
        """
        if self.is_alive():
            count = 0
            alive_count = 0
            for emitter in self.emitters:
                new_count = emitter(dt)
                if not _indicates_dead_emitter(new_count):
                    count += new_count
                    alive_count += 1
            self._alive_count = alive_count
            if count or self.is_alive():
                return count
            else:
                return OUT_OF_PARTICLES
        else:
            return OUT_OF_PARTICLES

    def is_alive(self):
        """Is this still expected to produce particles in the future?"""
        return self._alive_count > 0


class EmitterConstantRate(object):
    """Emit new particles at a constant rate."""
    def __init__(self, rate, limit=None):
        """Emit `rate` particles / second, up to a total `limit` lifetime.

        Arguments:
            rate: Number of particles to emit per second.
            limit: Maximum number of particles to ever emit.
        """
        self.rate = rate
        self.particles_left = limit
        self.remainder = 0

    def __call__(self, dt):
        """How many particles to emit in `dt` seconds.

        Arguments:
            dt: Elapsed time in seconds.
        """
        if self._has_limit and not self.particles_left:
            return OUT_OF_PARTICLES

        exact = self.remainder + dt * self.rate
        count = int(exact)
        self.remainder = exact - count
        if self._has_limit:
            count = min((count, self.particles_left))
            self.particles_left -= count

        return count

    @property
    def _has_limit(self):
        return self.particles_left is not None


class EmitterBurst(object):
    """Emit bursts of particles."""
    def __init__(self, counts_and_delays):
        """Set up the burst.

        Arguments:
            counts_and_delays (iterable): this is an iterable that
                yields pairs of (particle count, delay in seconds) allowing
                you to define multiple bursts, bursts after delays, etc.
        """
        self.iter_counts_and_delays = iter(counts_and_delays)
        self.delay = 0
        self.count = 0

    @classmethod
    def single(cls, count, delay=0):
        """Perform one single burst on particle system creation, launching
        count particles.

        Arguments:
            count: Number of particles to emit.
            delay: Time to wait before burst.
        """
        return cls([(count, delay)])

    @classmethod
    def repeat(cls, count, period, delay=0):
        """Perform regular, repeated bursts.

        Arguments:
            count: Number of particles to emit.
            period: Time to wait between bursts.
            delay: Time to wait before first burst.
        """
        return cls(
            itertools.chain(
                ((count, delay),),
                itertools.repeat((count, period)),
            )
        )

    @property
    def is_alive(self):
        """Whether it's still creating bursts in the future."""
        return self.delay < INFINITY

    def __call__(self, dt):
        """How many particles to emit in `dt` seconds.

        Arguments:
            dt: Elapsed time in seconds.
        """
        count = 0
        while dt >= self.delay:
            count += self.count
            dt -= self.delay
            self.count, self.delay = next(
                self.iter_counts_and_delays,
                (0, INFINITY)
            )
        self.delay -= dt
        if count or self.is_alive:
            return count
        else:
            return OUT_OF_PARTICLES


class PhysicsComposite(object):
    """Perform multiple physics operations in one call."""
    def __init__(self, *physics):
        """Create composite of multiple physics rules.

        Arguments:
            physics (list of physics callables): Physics callables, all of
                which will be applied to particles each frame.
        """
        self.physics = list(physics)

    def add(self, *physics):
        """Add a new physics function.

        Arguments:
            physics (list of physics callables): Physics callables, all of
                which will be applied to particles each frame.
        """
        self.physics.extend(physics)
        return self

    def __call__(self, dt, particle):
        """Do all physics updates for `dt` seconds on given particle.

        Arguments:
            dt (float): Elapsed seconds
            particle (Particle): Particle to update.
        """
        for physics in self.physics:
            physics(dt, particle)


class PhysicsInertia(object):
    """Perform inertia on a particle.

    Basically this will add the particle's velocity to position & expend
    the particle's life.
    """
    def __call__(self, dt, particle):
        """Perform inertia updates for `dt` seconds on given particle.

        Arguments:
            dt (float): Elapsed seconds
            particle (Particle): Particle to update.
        """
        particle.x += dt * particle.dx
        particle.y += dt * particle.dy
        particle.life -= dt


class PhysicsJitter(object):
    """Add randomness to particle motion.

    Randomness can be applied to particle coordinates, velocity, or life,
    which is cool.
    """

    ATTRS = 'x y dx dy life'.split()

    def __init__(self, x=0, y=0, dx=0, dy=0, life=0, jitter=None):
        """Define randomness to add.

        Arguments:
            x (float): scale to add x-jitter
            y (float): scale to add y-jitter
            dx (float): scale to add dx-jitter
            dy (float): scale to add dy-jitter
            life (float): scale to add life-jitter
            jitter (callable): random function, taking time delta, returning
                a random float multiplied by each attribute.
        """
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.life = life
        self.jitter = jitter or self.brownian

    @classmethod
    def brownian(cls, dt):
        return random.gauss(0, math.sqrt(dt))

    def __call__(self, dt, particle):
        """Perform jitter for `dt` seconds on given particle.

        Arguments:
            dt (float): Elapsed seconds
            particle (Particle): Particle to update.
        """
        # Jitter might be expensive so we only use it where necessary
        for attr in self.ATTRS:
            value = getattr(self, attr)
            if value:
                jitter = self.jitter(dt)
                # particle.attr += value * jitter
                pvalue = getattr(particle, attr)
                setattr(particle, attr, pvalue + value * jitter)


class PhysicsKick(object):
    """Kick a particle by adding whatever values to its attributes.

    This is intended for use in launchers, you can define an initial
    velocity for new particles, etc.
    """
    def __init__(self, x=0, y=0, dx=0, dy=0, life=0):
        """Set up values to add to kick particle.

        Arguments:
            x (float): value to add to x per second
            y (float): value to add to y per second
            dx (float): value to add to dx per second
            dy (float): value to add to dy per second
            life (float): value to add to life per second
        """
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.life = life

    def __call__(self, dt, particle):
        """Perform kick for `dt` seconds on given particle.

        Arguments:
            dt (float): Elapsed seconds
            particle (Particle): Particle to update.
        """
        particle.x += dt * self.x
        particle.y += dt * self.y
        particle.dx += dt * self.dx
        particle.dy += dt * self.dy
        particle.life += dt * self.life


class PhysicsAcceleration(object):
    """Constant acceleration field.

    This is kind of a special case of Kick, for only dx & dy, but
    we keep it separate for clarity.
    """
    def __init__(self, ax, ay):
        """Set up acceleration field.

        Arguments:
            ax (float): Acceleration in positive X direction,
                in pixels / sec / sec.
            ay (float): Acceleration in positive Y direction,
                in pixels / sec / sec.
        """
        self.ax = ax
        self.ay = ay

    @classmethod
    def radial(cls, r, theta):
        """Provide a radial acceleration.

         Arguments:
             r (float): speed in pixels per second (per second)
             theta (float): angle in degrees (0 = +X axis, 90 = +Y axis)
         """
        radians = math.radians(theta)
        ax = r * math.cos(radians)
        ay = r * math.sin(radians)
        return cls(ax=ax, ay=ay)

    def __call__(self, dt, particle):
        """Perform acceleration for `dt` seconds on given particle.

        Arguments:
            dt (float): Elapsed seconds
            particle (Particle): Particle to update.
        """
        particle.dx += dt * self.ax
        particle.dy += dt * self.ay


class ArtistSimple(object):
    """Artist that simply draws an image at each particle position."""
    def __init__(self, image, origin=CENTER, special_flags=0):
        """Initialize the artist.

        Arguments:
            image (pygame.Surface): Image to draw for each particle.
            special_flags (int): blit flags determining blit mode (normal,
                additive, subtractive, etc)
        """
        self.image = image
        if callable(origin):
            self.origin = origin(self.image)
        else:
            self.origin = origin
        self.special_flags = special_flags

    def __call__(self, surface, particle):
        """Draw given particle on given surface.

        Arguments:
            surface (pygame.Surface): Surface to draw on
            particle (Particle): Particle to draw
        """
        x = int(particle.x - self.origin[0])
        y = int(particle.y - self.origin[1])
        surface.blit(self.image, (x, y), special_flags=self.special_flags)


class ArtistFadeOverlay(object):
    """Artist that draws an image, fading between different tints based on
    the particle's lifetime."""
    def __init__(self, image, origin, tints,
                 blit_flags=0, tint_flags=pygame.BLEND_RGBA_MULT):
        """Initialize the artist.

        Arguments:
            image (pygame.Surface): Image to draw for each particle.
            tints (list of 4-tuple colors): Contains multiple 4-tuples,
                each of which indicates an RGBA tint to apply to the image.
                There should be at least two tints: if two are provided,
                the image will fade from the first tint to the second
                over its lifetime (determined by
                `particle.life / particle.initial_life`).  If more than two
                are given, particle will fade between different tints by
                dividing lifetime equally; so for 4 tints, particle will spend
                1/3 its lifetime between the first two, 1/3 between
                the middle two, and 1/3 between the last two.
            blit_flags (int): special flags determining blit mode (normal,
                additive, subtractive, etc) the tinted image will be applied
                to the surface. By default this is normal -- apply the
                image while respecting alpha, etc.)
            tint_flags (int): special flags determining blit mode (normal,
                additive, subtractive, etc) the tint will be applied to
                the image. By default this is multiplicative.
        """
        self.image = image
        try:
            self.image = self.image.convert_alpha()
        except:  # pragma: no cover
            pass

        if callable(origin):
            self.origin = origin(self.image)
        else:
            self.origin = origin
        self.tints = tints
        self.blit_flags = blit_flags
        self.tint_flags = tint_flags

    def __call__(self, surface, particle):
        """Draw given particle on given surface.

        Arguments:
            surface (pygame.Surface): Surface to draw on
            particle (Particle): Particle to draw
        """
        overlay = self.image.copy()
        tint = self.calculate_tint(particle.life, particle.initial_life)

        overlay.fill(tint, special_flags=self.tint_flags)
        x = int(particle.x - self.origin[0])
        y = int(particle.y - self.origin[1])
        surface.blit(overlay, (x, y), special_flags=self.blit_flags)

    def calculate_tint(self, life, initial_life):
        """Calculate tint to apply to particle."""
        life_fraction = (initial_life - life) * 1.0 / initial_life
        if life_fraction > 1:
            life_fraction = 1
        elif life_fraction < 0:
            life_fraction = 0
        index = life_fraction * (len(self.tints) - 1)
        floor = int(index)
        remainder = index - floor
        if remainder > 0:
            tint = tuple([
                int(round(a + (b - a) * remainder))
                for a, b in zip(self.tints[floor], self.tints[floor + 1])
            ])
        else:
            tint = self.tints[floor]
        print(tint)
        return tint


def _indicates_dead_emitter(desired_particle_count):
    return desired_particle_count is None or \
        desired_particle_count == OUT_OF_PARTICLES or \
        desired_particle_count < 0
